from .map_events import map_event_types
from .base_event import BaseItemEvent
from .item_events import ItemStateEvent, ItemStateChangedEvent, ItemCommandEvent, ItemAddedEvent,\
    ItemUpdatedEvent, ItemRemovedEvent, ItemStatePredictedEvent

EVENT_LIST = [ItemStateEvent, ItemStateChangedEvent, ItemCommandEvent, ItemAddedEvent,
              ItemUpdatedEvent, ItemRemovedEvent, ItemStatePredictedEvent]

__event_lookup = { k.__name__ : k for k in EVENT_LIST}

def get_event(_in_dict : dict) -> BaseItemEvent:
    event_type = _in_dict['type']
    try:
        return __event_lookup[event_type](_in_dict)
    except KeyError:
        raise ValueError(f'Unknown Event: {event_type:s}')



import typing
import datetime

from .scheduled_cb import ScheduledCallback

TYPING_TIME = typing.Union[datetime.time, datetime.timedelta, None]


class ReoccurringScheduledCallback(ScheduledCallback):

    CALL_ONCE = False

    def __init__(self, time: TYPING_TIME, interval: typing.Union[int, datetime.timedelta], callback, *args, **kwargs):
        assert isinstance(time, (datetime.time, datetime.timedelta)) or time is None, f'{time} ({type(time)})'
        super().__init__(time, callback, *args, **kwargs)

        if isinstance(interval, int):
            interval = datetime.timedelta(seconds=interval)
        assert isinstance(interval, datetime.timedelta), type(interval)
        self.time_interval = interval

    def check_due(self, now: datetime):
        super().check_due(now)
        if self.is_due:
            self.next_call += self.time_interval


class DayOfWeekScheduledCallback(ScheduledCallback):

    CALL_ONCE = False

    def __init__(self, time: TYPING_TIME, weekdays: typing.List[int], callback, *args, **kwargs):
        assert isinstance(time, (datetime.time, datetime.timedelta)) or time is None, f'{time} ({type(time)})'
        super().__init__(time, callback, *args, **kwargs)

        assert weekdays, 'please specify weekdays'
        assert isinstance(weekdays, list)
        for k in weekdays:
            assert 1 <= k <= 7, k
        self.weekdays = weekdays

        # shedule for next day if the date_time is in the past or date does not match weekdays
        while not self.next_call.isoweekday() in self.weekdays:
            self.next_call += datetime.timedelta(days=1)

    def check_due(self, now: datetime):
        super().check_due(now)
        if self.is_due:
            self.next_call += datetime.timedelta(days=1)
            while not self.next_call.isoweekday() in self.weekdays:
                self.next_call += datetime.timedelta(days=1)


class WorkdayScheduledCallback(DayOfWeekScheduledCallback):
    def __init__(self, time: TYPING_TIME, callback, *args, **kwargs):
        super().__init__(time, [1, 2, 3, 4, 5], callback, *args, **kwargs)


class WeekendScheduledCallback(DayOfWeekScheduledCallback):
    def __init__(self, time: TYPING_TIME, callback, *args, **kwargs):
        super().__init__(time, [6, 7], callback, *args, **kwargs)



import typing
from datetime import datetime, timedelta, time

from HABApp.core.worker import WrappedFunction

TYPING_DATE_TIME = typing.Union[None, datetime, timedelta, time]


class ScheduledCallback:

    VALIDATE_CALLBACK_DATETIME = True
    CALL_ONCE = True

    def __init__(self, date_time: TYPING_DATE_TIME, callback, *args, **kwargs):

        # next time the callback will be executed
        __now = datetime.now()

        if date_time is None:
            # If we don't specify a datetime we start it now
            date_time = __now
        elif isinstance(date_time, timedelta):
            # if it is a timedelta add it to now to easily speciy points in the future
            date_time = __now + date_time
        elif isinstance(date_time, time):
            # if it is a time object it specifies a time of day.
            date_time = __now.replace(hour=date_time.hour, minute=date_time.minute, second=date_time.second)
            if date_time < __now:
                date_time += timedelta(days=1)

        assert isinstance(date_time, datetime), type(date_time)

        if ScheduledCallback.VALIDATE_CALLBACK_DATETIME:
            assert date_time >= __now, f'Time for callback must be in the future!\nTime: {date_time}\nNow : {__now}'

        assert isinstance(callback, WrappedFunction), type(callback)
        self._callback = callback
        self._args = args
        self._kwargs = kwargs

        self.is_due = False
        self.is_finished = False

        self.next_call = date_time

    def get_next_call(self):
        "Return the next execution timestamp"
        return self.next_call

    def check_due(self, now : datetime):

        self.is_due = True if self.next_call <= now else False
        if self.is_finished:
            self.is_due = False

        return self.is_due

    def execute(self):
        if not self.is_due or self.is_finished:
            return False

        self._callback.submit(*self._args, **self._kwargs)

        if self.__class__.CALL_ONCE:
            self.is_finished = True
        return True

    def cancel(self):
        self.is_finished = True



from ..runtime import Runtime


class RuleParameter:
    def __init__(self, runtime, filename, *keys):
        assert isinstance(runtime, Runtime)
        self.__runtime = runtime
        self.filename = filename
        self.keys = keys

        # as a convenience try to create the file and the file structure
        self.__runtime.rule_params.add_param(self.filename, *self.keys)

    def get_value(self):
        return self.__runtime.rule_params.get_param(self.filename, *self.keys)

    def __eq__(self, other):
        return self.get_value() == other

    def __ne__(self, other):
        return self.get_value() != other

    def __lt__(self, other):
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self.get_value() < other

    def __le__(self, other):
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self.get_value() <= other

    def __ge__(self, other):
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self.get_value() >= other

    def __gt__(self, other):
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self.get_value() > other

    def __repr__(self):
        return f'<RuleParameter file: {self.filename}, keys: {self.keys}, ' \
               f'value: {self.get_value()}'



import HABApp.util

from HABApp.rule import Rule
from HABApp.runtime import Runtime

import HABApp.core as core

import HABApp.config
import HABApp.mqtt
import HABApp.openhab
import HABApp.openhab.events as events



import datetime
import re

# function which generates the datetime obj
_TIME_FUNC = datetime.datetime.now

# names of weekdays in local language
_WEEKDAY_NAMES = {i: datetime.date(2001, 1, i + 1).strftime('%A')[:3] for i in range(0, 7)}
_DAYS = {datetime.date(2001, 1, i + 1).strftime('%A'): i for i in range(0, 7)}
# _DAYS.update(_WEEKDAY_NAMES)

# abreviations in German and English
_DAYS.update({"Mo": 0, "Di": 1, "Mi": 2, "Do": 3, "Fr": 4, "Sa": 5, "So": 6})
_DAYS.update({"Mon": 0, "Tue": 1, "Wed": 2, "Thu": 3, "Fri": 4, "Sat": 5, "Sun": 6})
_DAYS = {k.lower(): v for k, v in _DAYS.items()}

_RE_VALID = re.compile(r'^((?:\w+,)*)((?:\d+:\d+:?\d*)(?:-\d+:\d+:?\d*)?)$')
_RE_DAYS = re.compile(r'(\w+),')
_RE_TIMES = re.compile(r'^(\d+:\d+:?\d*)(?:-(\d+:\d+:?\d*))?$')
_RE_TIME = re.compile(r'^(\d+):(\d+)(?::(\d+))?$')


class TimeFrame(object):

    def __init__(self, timespan, log_function=None):
        """
        Creates TimeFrame object.

        :param timespan: Description of time span. Omitting second time means until end of day (23:59:59).
        E.g. "mon,8:30:00-9:31", "weekends,22:30:00", 'weekdays,8:30:00-9:30:00", "mon,tue,8:30:00-9:31:30"
        :param log_function: function for logging which takes 1 parameter
        """
        assert isinstance(timespan, str)
        if log_function is not None:
            assert callable(log_function)
        desc = timespan.lower().replace(' ', '')

        self.time_start = None  # type: tuple
        self.time_stop = None  # type: tuple
        self.weekdays = []  # type: list[int]

        self.__log_func = log_function

        m = _RE_VALID.search(desc)
        if not m or len(m.groups()) != 2:
            raise ValueError('Invalid format!')

        self.__get_days(m.group(1))
        self.__get_time(m.group(2))

        self.__check_time()

        self.desc_weeks = ''
        for day in self.weekdays:
            self.desc_weeks += _WEEKDAY_NAMES[day] + ","
        self.desc_time = '{:2d}:{:02d}:{:02d}-{:2d}:{:02d}:{:02d}'.format(self.time_start[0], self.time_start[1],
                                                                          self.time_start[2],
                                                                          self.time_stop[0], self.time_stop[1],
                                                                          self.time_stop[2])

    def __check_time(self):
        ts = [self.time_start, self.time_stop]
        for t in ts:
            if not 0 <= t[0] <= 23:
                raise ValueError('Hour must be 0 - 23!')
            if not 0 <= t[1] <= 59 or not 0 <= t[2] <= 59:
                raise ValueError('Minutes and seconds must be 0 - 59!')

        t1 = self.time_start[0] * 3600 + self.time_start[1] * 60 + self.time_start[2]
        t2 = self.time_stop[0] * 3600 + self.time_stop[1] * 60 + self.time_stop[2]

        if t2 <= t1:
            raise ValueError('Stop must be after start!\nStart: {}\nStop : {}'.format(self.time_start, self.time_stop))

    def __get_days(self, desc):
        # no day is all days!
        if desc == "":
            for i in range(7):
                self.weekdays.append(i)
            return

        m = _RE_DAYS.findall(desc)
        if not m:
            raise ValueError('No day specified in "{}"'.format(desc))

        def add_day(x):
            if x not in self.weekdays:
                self.weekdays.append(x)

        for d in m:
            if d == 'weekdays' or d.startswith('werktag'):
                for i in range(5):
                    add_day(i)
            elif d == 'weekends' or d == 'wochenende':
                add_day(5)
                add_day(6)
            else:
                if d in _DAYS:
                    add_day(_DAYS[d])
                else:
                    raise ValueError('Value "{:s}" ist not a valid day! (Allowed: {})'.format(d, list(_DAYS.keys())))

    def __get_time(self, desc):
        m = _RE_TIMES.search(desc)
        assert len(m.groups()) == 2, m.groups()

        t1 = _RE_TIME.search(m.group(1))
        l1 = [int(k) if k is not None else 0 for k in t1.groups()]

        # fehlt der zweite Eintrag dann gilt es bis Tagesende
        if m.group(2) is not None:
            t2 = _RE_TIME.search(m.group(2))
            l2 = [int(k) if k is not None else 0 for k in t2.groups()]
        else:
            l2 = [23, 59, 59]

        assert len(l1) == 3, '{}'.format(l1)
        assert len(l2) == 3, '{}'.format(l2)

        self.time_start = tuple(l1)
        self.time_stop = tuple(l2)

    def __log(self, _in):
        if self.__log_func is not None:
            self.__log_func(_in)

    def is_now(self):
        """
        Compares the defined time span to now

        :return: True if now is in time span, else False
        """

        date_now = _TIME_FUNC()
        assert isinstance(date_now, datetime.datetime), type(date_now)

        self.__log('Def : {:s}{:s}'.format(self.desc_weeks, self.desc_time))
        __now_str = 'Now :{:>{width}s},{:2d}:{:02d}:{:02d}'.format(_WEEKDAY_NAMES[date_now.weekday()], date_now.hour,
                                                                   date_now.minute, date_now.second,
                                                                   width=len(self.desc_weeks))

        day_now = date_now.weekday()
        if day_now not in self.weekdays:
            self.__log(__now_str + ' (False)')
            return False

        start = 3600 * self.time_start[0] + 60 * self.time_start[1] + self.time_start[2]
        stop = 3600 * self.time_stop[0] + 60 * self.time_stop[1] + self.time_stop[2]
        now = 3600 * date_now.hour + 60 * date_now.minute + date_now.second

        if start <= now <= stop:
            self.__log(__now_str + ' (True)')
            return True
        else:
            self.__log(__now_str + ' (False)')
            return False



import HABApp.core

from .map_events import map_event_types
from .base_event import BaseItemEvent


class ItemStateEvent(BaseItemEvent, HABApp.core.ValueUpdateEvent):
    def __init__(self, _in_dict):
        super().__init__(_in_dict)

        # smarthome/items/NAME/state
        self.name = self._topic[16:-6]
        self.value = map_event_types(self._payload['type'], self._payload['value'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class ItemStateChangedEvent(BaseItemEvent, HABApp.core.ValueChangeEvent):
    def __init__(self, _in_dict):
        super().__init__(_in_dict)

        # smarthome/items/Ping/statechanged
        self.name = self._topic[16:-13]
        self.value = map_event_types(self._payload['type'], self._payload['value'])
        self.old_value = map_event_types(self._payload['oldType'], self._payload['oldValue'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}, old_value: {self.old_value}>'


class ItemCommandEvent(BaseItemEvent):
    def __init__(self, _in_dict):
        super().__init__(_in_dict)

        # smarthome/items/NAME/command
        self.name = self._topic[16:-8]
        self.value = map_event_types(self._payload['type'], self._payload['value'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class ItemAddedEvent(BaseItemEvent):
    def __init__(self, _in_dict):
        super().__init__(_in_dict)

        # {'topic': 'smarthome/items/NAME/added'
        # 'payload': '{"type":"Contact","name":"Test","tags":[],"groupNames":[]}'
        # 'type': 'ItemAddedEvent'}
        self.name = self._payload['name']
        self.type = self._payload['type']

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, type: {self.type}>'


class ItemUpdatedEvent(BaseItemEvent):
    def __init__(self, _in_dict):
        super().__init__(_in_dict)

        # smarthome/items/NAME/updated
        # 'payload': '[{"type":"Switch","name":"Test","tags":[],"groupNames":[]},
        #              {"type":"Contact","name":"Test","tags":[],"groupNames":[]}]',
        # 'type': 'ItemUpdatedEvent'
        self.name = self._topic[16:-8]
        self.type = self._payload[1]['type']

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, type: {self.type}>'


class ItemRemovedEvent(BaseItemEvent):
    def __init__(self, _in_dict):
        super().__init__(_in_dict)

        # smarthome/items/Test/removed
        self.name = self._topic[16:-8]

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}>'


class ItemStatePredictedEvent(BaseItemEvent):
    def __init__(self, _in_dict):
        super().__init__(_in_dict)

        # 'smarthome/items/NAME/statepredicted'
        self.name = self._topic[16:-15]
        self.value = map_event_types(self._payload['predictedType'], self._payload['predictedValue'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'



import asyncio
import logging
import time
import traceback

import HABApp
import HABApp.core
import HABApp.openhab.events
from HABApp.openhab.events import get_event
from HABApp.util import PrintException
from .http_connection import HttpConnection, HttpConnectionEventHandler
from .oh_interface import OpenhabInterface

log = logging.getLogger('HABApp.openhab.Connection')



class OpenhabConnection(HttpConnectionEventHandler):

    def __init__(self, parent):
        assert isinstance(parent, HABApp.Runtime)
        self.runtime: HABApp.Runtime = parent

        self.connection = HttpConnection(self, self.runtime.config)
        self.interface = OpenhabInterface(self.connection, openhab_config=self.runtime.config.openhab)

        self.__ping_sent = 0
        self.__ping_received = 0

        # Add the ping listener, this works because connect is the last step
        if self.runtime.config.openhab.ping.enabled:
            listener = HABApp.core.EventListener(
                self.runtime.config.openhab.ping.item,
                HABApp.core.WrappedFunction(self.ping_received),
                HABApp.openhab.events.ItemStateEvent
            )
            HABApp.core.Events.add_listener(listener)

        self.runtime.shutdown.register_func(self.shutdown)

        # todo: currently this does not work
        # # reload config
        # self.runtime.config.openhab.connection.subscribe_for_changes(
        #     lambda : asyncio.run_coroutine_threadsafe(self.__create_session(), self.runtime.loop).result()
        # )

        self.__async_sse: asyncio.Future = None
        self.__async_items: asyncio.Future = None

    def on_connected(self):
        # Start SSE Processing
        self.__async_sse = asyncio.ensure_future(self.connection.async_process_sse_events())

        # Read all Items
        self.__async_items = asyncio.ensure_future(self.update_all_items())


    def on_disconnected(self):
        if self.__async_sse:
            if not self.__async_sse.done():
                self.__async_sse.cancel()

        if self.__async_items:
            if not self.__async_items.done():
                self.__async_items.cancel()

    async def start(self):
        await self.connection.try_connect()

    def shutdown(self):
        self.on_disconnected()
        self.connection.cancel_connect()

    @PrintException
    def ping_received(self, event):
        self.__ping_received = time.time()

    @PrintException
    async def async_ping(self):

        if not self.runtime.config.openhab.ping.enabled:
            return None

        log.debug('Started ping')
        while self.runtime.config.openhab.ping.enabled:
            self.interface.post_update(
                self.runtime.config.openhab.ping.item,
                f'{(self.__ping_received - self.__ping_sent) * 1000:.1f}' if self.__ping_received else '0'
            )

            self.__ping_sent = time.time()
            await asyncio.sleep(10)



    def on_sse_event(self, event: dict):

        # Lookup corresponding OpenHAB event
        event = get_event(event)

        # Events which change the ItemRegistry
        if isinstance(event, (HABApp.openhab.events.ItemAddedEvent, HABApp.openhab.events.ItemUpdatedEvent)):
            item = HABApp.openhab.map_items(event.name, event.type, 'NULL')
            HABApp.core.Items.set_item(item)
        elif isinstance(event, HABApp.openhab.events.ItemRemovedEvent):
            HABApp.core.Items.pop_item(event.name)

        # Send Event to Event Bus
        HABApp.core.Events.post_event(event.name, event)


    @PrintException
    async def update_all_items(self) -> int:

        try:
            data = await self.connection.async_get_items()
            if data is None:
                return None

            found_items = len(data)
            for _dict in data:
                __item = HABApp.openhab.map_items(_dict['name'], _dict['type'], _dict['state'])
                HABApp.core.Items.set_item(__item)

            # remove items which are no longer available
            ist = set(HABApp.core.Items.items.keys())
            soll = {k['name'] for k in data}
            for k in ist - soll:
                HABApp.core.Items.pop_item(k)

            log.info(f'Updated {found_items:d} items')
            return found_items
        except Exception as e:
            log.error(e)
            for l in traceback.format_exc().splitlines():
                log.error(l)
            return 0



import logging
import ujson

import paho.mqtt.client as mqtt

import HABApp
from HABApp.util import PrintException

log = logging.getLogger('HABApp.mqtt.connection')
log_msg = logging.getLogger('HABApp.Events.mqtt')


class MqttInterface:
    def __init__(self, connection):
        self.__connection: MqttConnection = connection

    def publish(self, topic: str, payload, qos=None, retain=None):
        self.__connection.publish(topic, payload, qos, retain)


class MqttConnection:
    def __init__(self, parent):
        assert isinstance(parent, HABApp.Runtime)
        self.runtime: HABApp.Runtime = parent

        self.loop_started = False
        self.connected = False

        self.client: mqtt.Client = None

        self.subscriptions = []
        self.value_cache = {}

        # config changes
        self.runtime.config.mqtt.subscribe.subscribe_for_changes(self.subscription_changed)
        self.runtime.config.mqtt.connection.subscribe_for_changes(self.connect)

        # shutdown
        self.runtime.shutdown.register_func(self.disconnect)

        self.interface: MqttInterface = MqttInterface(self)

    def connect(self):
        if not self.runtime.config.mqtt.connection.host:
            log.info('MQTT disabled')
            self.disconnect()
            return None

        if self.connected:
            log.info('disconnecting')
            self.client.disconnect()
            self.connected = False

        self.client = mqtt.Client(
            client_id=self.runtime.config.mqtt.connection.client_id,
            clean_session=False
        )

        if self.runtime.config.mqtt.connection.tls:
            self.client.tls_set()

        if self.runtime.config.mqtt.connection.tls_insecure:
            log.warning('Verification of server hostname in server certificate disabled!')
            log.warning('Use this only for testing, not for a real system!')
            self.client.tls_insecure_set(True)

        # set user if required
        user = self.runtime.config.mqtt.connection.user
        pw = self.runtime.config.mqtt.connection.password
        if user:
            self.client.username_pw_set(
                user,
                pw if pw else None
            )

        # setup callbacks
        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect
        self.client.on_message = self.process_msg

        self.client.connect_async(
            self.runtime.config.mqtt.connection.host,
            port=self.runtime.config.mqtt.connection.port,
            keepalive=60
        )

        if not self.loop_started:
            self.client.loop_start()
        self.loop_started = True

    @PrintException
    def subscription_changed(self):
        if not self.connected:
            return None

        if self.subscriptions:
            unsubscribe = [k[0] for k in self.subscriptions]
            for t in unsubscribe:
                log.debug(f'Unsubscribing from "{t}"')
            self.client.unsubscribe(unsubscribe)

        topics = self.runtime.config.mqtt.subscribe.topics
        default_qos = self.runtime.config.mqtt.subscribe.qos
        self.subscriptions = [(topic, qos if qos is not None else default_qos) for topic, qos in topics]
        for topic, qos in self.subscriptions:
            log.debug(f'Subscribing to "{topic}" (QoS {qos:d})')
        self.client.subscribe(self.subscriptions)

    @PrintException
    def on_connect(self, client, userdata, flags, rc):
        log.log(logging.INFO if not rc else logging.ERROR, mqtt.connack_string(rc))
        if rc:
            return None
        self.connected = True

        self.value_cache.clear()
        self.subscriptions.clear()
        self.subscription_changed()

    @PrintException
    def on_disconnect(self, client, userdata, rc):
        log.log(logging.INFO if not rc else logging.ERROR, "Unexpected disconnection")
        self.connected = False

    @PrintException
    def process_msg(self, client, userdata, message: mqtt.MQTTMessage):
        topic = message.topic
        payload = message.payload.decode("utf-8")

        if log_msg.isEnabledFor(logging.DEBUG):
            log_msg._log(logging.DEBUG, f'{topic} ({message.qos}): {payload}', [])

        # load json dict and list
        if payload.startswith('{') and payload.endswith('}') or payload.startswith('[') and payload.endswith(']'):
            try:
                payload = ujson.loads(payload)
            except ValueError:
                pass
        else:
            # try to cast to int/float
            try:
                payload = int(payload)
            except ValueError:
                try:
                    payload = float(payload)
                except ValueError:
                    pass

        __was = self.value_cache.get(topic, None)
        self.value_cache[topic] = payload

        HABApp.core.Events.post_event(topic, HABApp.core.ValueUpdateEvent(topic, payload))
        if __was is not None and __was != payload:
            HABApp.core.Events.post_event(topic, HABApp.core.ValueChangeEvent(topic, payload, __was))

    def publish(self, topic, payload, qos, retain) -> mqtt.MQTTMessageInfo:

        if not self.connected:
            raise ConnectionError(f'Mqtt client not connected')

        if qos is None:
            qos = self.runtime.config.mqtt.publish.qos
        if retain is None:
            retain = self.runtime.config.mqtt.publish.retain

        info = self.client.publish(topic, payload, qos, retain)
        if info.rc != mqtt.MQTT_ERR_SUCCESS:
            log.error(f'Could not publish to "{topic}": {mqtt.error_string(info.rc)}')
        return info

    def disconnect(self):
        if self.connected:
            self.client.disconnect()
            self.connected = False
        if self.loop_started:
            self.client.loop_stop()
            self.loop_started = False

        self.client = None




def get_default_logfile() -> str:
    return """version : 1


formatters:
  HABApp_format:
    format: '[%(asctime)s] [%(name)25s] %(levelname)8s | %(message)s'


handlers:
  HABApp_default:
    class: logging.handlers.RotatingFileHandler
    filename: 'HABApp.log'
    maxBytes: 10_485_760
    backupCount: '3'

    formatter: HABApp_format
    level: DEBUG

  EventFile:
    class: logging.handlers.RotatingFileHandler
    filename: 'events.log'
    maxBytes: 10_485_760
    backupCount: 3

    formatter: HABApp_format
    level: DEBUG

  BufferEventFile:
    class: logging.handlers.MemoryHandler
    capacity: 10
    formatter: HABApp_format
    target: EventFile
    level: DEBUG


loggers:
  HABApp:
    level: INFO
    handlers:
      - HABApp_default
    propagate: False

  HABApp.Events:
    level: INFO
    handlers:
      - BufferEventFile
    propagate: False

"""



from HABApp.core.items import Item


class SwitchItem(Item):
    ON = 'ON'
    OFF = 'OFF'

    @classmethod
    def from_str(self, name, value):
        item = SwitchItem(name=name)
        item.set_state(value)
        return item

    def set_state(self, new_state):
        if new_state is not None and new_state != SwitchItem.ON and new_state != SwitchItem.OFF:
            raise ValueError(f'Invalid value for SwitchItem {self.name}: {new_state}')
        super().set_state(new_state)

    def is_on(self):
        return True if self.state == SwitchItem.ON else False

    def is_off(self):
        return True if self.state == SwitchItem.OFF else False

    def __str__(self):
        return self.state

    def __eq__(self, other):
        if isinstance(other, SwitchItem):
            return self.state == other.state
        elif isinstance(other, str):
            return self.state == other
        elif isinstance(other, int):
            if other and self.is_on():
                return True
            if not other and self.is_off():
                return True
            return False

        return NotImplemented

    def __ne__(self, other):
        res = self.__eq__(other)
        if res is NotImplemented:
            return res
        return not res



import re

from voluptuous import Invalid

from .configentry import ConfigEntry, ConfigEntryContainer


def TimeZoneValidator(msg=None):
    __re = re.compile(r'[+-]\d{4}')

    def f(v):
        v = str(v)
        if __re.fullmatch(v):
            return v
        else:
            raise Invalid(msg or (f"incorrect timezone ({v})! Example: +1000 or -1030"))

    return f


class Ping(ConfigEntry):
    def __init__(self):
        super().__init__()
        self.enabled = False
        self.item = ''
        self.interval = 10


class General(ConfigEntry):
    def __init__(self):
        super().__init__()
        self.timezone = '+1000'
        self._entry_validators['timezone'] = TimeZoneValidator()
        self.listen_only = False


class Connection(ConfigEntry):
    def __init__(self):
        super().__init__()
        self.host = 'localhost'
        self.port = 8080
        self.user = ''
        self.password = ''

        self._entry_kwargs['user'] = {'default': ''}
        self._entry_kwargs['password'] = {'default': ''}


class Openhab(ConfigEntryContainer):
    def __init__(self):
        self.ping = Ping()
        self.connection = Connection()
        self.general = General()



from pathlib import Path

from watchdog.events import FileSystemEventHandler, FileMovedEvent, FileCreatedEvent, \
    FileDeletedEvent, FileModifiedEvent


class FileEventTarget:
    def add_file(self, path: Path):
        raise NotImplementedError()

    def reload_file(self, path: Path):
        raise NotImplementedError()

    def remove_file(self, path: Path):
        raise NotImplementedError()


class SimpleFileEventHandler(FileSystemEventHandler):
    def __init__(self, event_target, file_ending: str, worker_factory=None):
        assert isinstance(event_target, FileEventTarget), type(event_target)
        self.target = event_target

        assert isinstance(file_ending, str), type(file_ending)
        self.__file_ending = file_ending

        # Possibility to use a wrapper to load files
        # do not reuse an instantiated WrappedFunction because it will throw errors in the traceback module
        self.__worker_factory = worker_factory

    def _get_func(self, func):
        if self.__worker_factory is None:
            return func
        return self.__worker_factory(func)

    def on_deleted(self, event):
        if not isinstance(event, FileDeletedEvent):
            return None

        if not event.src_path.endswith(self.__file_ending):
            return None

        self._get_func(self.target.remove_file)(Path(event.src_path))

    def on_modified(self, event):
        if not isinstance(event, FileModifiedEvent):
            return None

        if not event.src_path.endswith(self.__file_ending):
            return None

        self._get_func(self.target.reload_file)(Path(event.src_path))

    def on_created(self, event):
        if not isinstance(event, FileCreatedEvent):
            return None

        if not event.src_path.endswith(self.__file_ending):
            return None

        self._get_func(self.target.add_file)(Path(event.src_path))

    def on_moved(self, event):
        if not isinstance(event, FileMovedEvent):
            return None

        if event.src_path.endswith(self.__file_ending):
            self._get_func(self.target.remove_file)(Path(event.src_path))

        if event.dest_path.endswith(self.__file_ending):
            self._get_func(self.target.add_file)(Path(event.dest_path))



import logging
import threading
import traceback
import typing
from pathlib import Path

import ruamel.yaml

import HABApp
from HABApp.runtime import FileEventTarget
from HABApp.util import PrintException

log = logging.getLogger('HABApp.RuleParameters')

_yml_setup = ruamel.yaml.YAML()
_yml_setup.default_flow_style = False
_yml_setup.default_style = False
_yml_setup.width = 1000000
_yml_setup.allow_unicode = True
_yml_setup.sort_base_mapping_type_on_output = False


class RuleParameters(FileEventTarget):

    @PrintException
    def __init__(self, parent):
        assert isinstance(parent, HABApp.Runtime)
        self.runtime = parent

        # serialize loading
        self.__lock = threading.Lock()

        self.params: typing.Dict[str, dict] = {}

        if not self.runtime.config.directories.param.is_dir():
            log.info(f'Parameter files disabled. Folder {self.runtime.config.directories.param} does not exist!')
            return

        # folder watcher
        self.runtime.folder_watcher.watch_folder(
            folder=self.runtime.config.directories.param,
            file_ending='.yml',
            event_target=self,
            worker_factory=lambda x : HABApp.core.WrappedFunction(x, logger=log).submit,
            watch_subfolders=False
        )

        # load all parameter files
        def load_all_files():
            for f in self.runtime.config.directories.param.glob('**/*.yml'):
                try:
                    self.add_file(f)
                except Exception as e:
                    log.error(e)
        HABApp.core.WrappedFunction(load_all_files, logger=log, name='Load all parameter files').submit()

    def reload_file(self, path: Path):
        self.add_file(path)

    def remove_file(self, path: Path):
        try:
            with self.__lock:
                self.params.pop(path.stem)
        except Exception:
            log.error(f"Could not remove parameters from {path.name}!")
            for l in traceback.format_exc().splitlines():
                log.error(l)
            return None

        log.debug(f'Removed params from {path.name}!')

    def add_file(self, path : Path):
        try:
            with self.__lock:
                with path.open( mode='r', encoding='utf-8') as file:
                    data = _yml_setup.load(file)
                if data is None:
                    data = {}
                self.params[path.stem] = data
        except Exception:
            log.error(f"Could not load params from {path.name}!")
            for l in traceback.format_exc().splitlines():
                log.error(l)
            return None

        log.debug(f'Loaded params from {path.name}!')

    def add_param(self, file, *keys):
        save = False

        if file not in self.params:
            save = True
            param = data = {}
        else:
            param = data = self.params[file]

        if keys:
            # Create structure
            for key in keys[:-1]:
                if key not in param:
                    param[key] = {}
                    save = True
                param = param[key]

            # Create value
            if keys[-1] not in param:
                param[keys[-1]] = 'ToDo'
                save = True

        if save:
            filename = self.runtime.config.directories.param / (file + '.yml')
            with self.__lock:
                log.info(f'Updated {filename}')
                with open(filename, 'w', encoding='utf-8') as file:
                    _yml_setup.dump(data, file)
        return None

    def get_param(self, file, *keys):
        try:
            param = self.params[file]
        except KeyError:
            raise FileNotFoundError(f'File {file}.yml not found in params folder!')

        # lookup parameter
        for key in keys:
            param = param[key]
        return param



import codecs
import logging
import logging.config
import time
from pathlib import Path
import sys

import ruamel.yaml
from voluptuous import MultipleInvalid, Schema

from HABApp.__version__ import __VERSION__
from ._conf_mqtt import Mqtt
from ._conf_openhab import Openhab
from .configentry import ConfigEntry
from .default_logfile import get_default_logfile

from HABApp.runtime import FileEventTarget

_yaml_param = ruamel.yaml.YAML(typ='safe')
_yaml_param.default_flow_style = False
_yaml_param.default_style = False
_yaml_param.width = 1000000
_yaml_param.allow_unicode = True
_yaml_param.sort_base_mapping_type_on_output = False


log = logging.getLogger('HABApp.Config')


class Directories(ConfigEntry):
    def __init__(self):
        super().__init__()
        self.logging: Path = 'log'
        self.rules: Path   = 'rules'
        self.lib: Path     = 'lib'
        self.param: Path   = 'param'


class Config(FileEventTarget):

    def __init__(self, runtime, config_folder : Path):

        import HABApp.runtime
        assert isinstance(runtime, HABApp.runtime.Runtime)
        self.__runtime = runtime

        assert isinstance(config_folder, Path)
        assert config_folder.is_dir(), config_folder
        self.folder_conf = config_folder
        self.file_conf_habapp  = self.folder_conf / 'config.yml'
        self.file_conf_logging = self.folder_conf / 'logging.yml'

        # these are the accessible config entries
        self.directories = Directories()
        self.mqtt = Mqtt()
        self.openhab = Openhab()

        # if the config does not exist it will be created
        self.__check_create_config()
        self.__check_create_logging()

        # folder watcher
        self.__runtime.folder_watcher.watch_folder(
            folder=self.folder_conf,
            file_ending='.yml',
            event_target=self
        )

        # Load Config initially
        self.first_start = True
        self.add_file(self.file_conf_habapp)
        self.add_file(self.file_conf_logging)
        self.first_start = False

    def add_file(self, path: Path):
        self.reload_file(path)

    def reload_file(self, path: Path):
        if path.name == 'config.yml':
            self.load_cfg()
        if path.name == 'logging.yml':
            self.load_log()

    def remove_file(self, path: Path):
        pass

    def __check_create_config(self):
        if self.file_conf_habapp.is_file():
            return None

        cfg = {}
        self.directories.insert_data(cfg)
        self.openhab.insert_data(cfg)
        self.mqtt.insert_data(cfg)

        print( f'Creating {self.file_conf_habapp.name} in {self.file_conf_habapp.parent}')
        with open(self.file_conf_habapp, 'w', encoding='utf-8') as file:
            _yaml_param.dump(cfg, file)

        time.sleep(0.1)
        return None


    def __check_create_logging(self):
        if self.file_conf_logging.is_file():
            return None

        print(f'Creating {self.file_conf_logging.name} in {self.file_conf_logging.parent}')
        with open(self.file_conf_logging, 'w', encoding='utf-8') as file:
            file.write(get_default_logfile())

        time.sleep(0.1)
        return None

    def load_cfg(self):
        # File has to exist - check because we also get FileDelete events
        if not self.file_conf_habapp.is_file():
            return

        with open( self.file_conf_habapp, 'r', encoding='utf-8') as file:
            cfg = _yaml_param.load(file)
        try:
            _s = {}
            self.directories.update_schema(_s)
            self.openhab.update_schema(_s)
            self.mqtt.update_schema(_s)
            cfg = Schema(_s)(cfg)
        except MultipleInvalid as e:
            log.error( f'Error loading config:')
            log.error( e)
            return

        self.directories.load_data(cfg)
        self.openhab.load_data(cfg)
        self.mqtt.load_data(cfg)

        # make Path absolute for all directory entries
        for k, v in self.directories.iter_entry():
            __entry  = Path(v)
            if not __entry.is_absolute():
                __entry = self.folder_conf / __entry
                self.directories.__dict__[k] = __entry.resolve()

        if not self.directories.logging.is_dir():
            print( f'Creating log-dir: {self.directories.logging}')
            self.directories.logging.mkdir()

        log.debug('Loaded HABApp config')

        # Set path for libraries
        if self.directories.lib.is_dir():
            lib_path = str(self.directories.lib)
            if lib_path not in sys.path:
                sys.path.insert(0, lib_path)
                log.debug( f'Added library folder "{lib_path}" to path')


    def load_log(self):
        # File has to exist - check because we also get FileDelete events
        if not self.file_conf_logging.is_file():
            return None

        with open(self.file_conf_logging, 'r', encoding='utf-8') as file:
            cfg = _yaml_param.load(file)

        # fix filenames
        for handler, handler_cfg in cfg.get('handlers', {}).items():

            # fix encoding for FileHandlers - we always log utf-8
            if 'file' in handler_cfg.get('class', '').lower():
                enc = handler_cfg.get('encoding', '')
                if enc != 'utf-8':
                    handler_cfg['encoding'] = 'utf-8'

            if 'filename' not in handler_cfg:
                continue

            # make Filenames absolute path in the log folder if not specified
            p = Path(handler_cfg['filename'])
            if not p.is_absolute():
                p = (self.directories.logging / p).resolve()
                handler_cfg['filename'] = str(p)

            # Delete old Log-Files on startup
            if self.first_start and p.is_file():
                try:
                    # default is utf-8 logging so we write BOM
                    with open(p, mode='wb') as f:
                        f.write(codecs.BOM_UTF8)
                finally:
                    pass

        # load prepared logging
        logging.config.dictConfig(cfg)
        log.debug('Loaded logging config')

        logging.getLogger('HABApp').info(f'HABApp Version {__VERSION__}')



import datetime

from HABApp.core import Events, Items, ValueNoChangeEvent, ValueNoUpdateEvent


class WatchedItem:
    def __init__(self, name, constant_time: int, watch_only_changes=False):
        assert isinstance(name, str)
        assert isinstance(constant_time, int)
        assert isinstance(watch_only_changes, bool)

        self.name: str = name
        self.duration_const = datetime.timedelta(seconds=constant_time)
        self.executed = False

        self.__watch_only_changes = watch_only_changes

        self.is_canceled = False

    def check(self, now):
        if self.is_canceled:
            return None

        try:
            item = Items.get_item( self.name)
        except KeyError:
            return None

        timestamp = item.last_change if self.__watch_only_changes else item.last_update
        duration = now - timestamp
        if duration < self.duration_const:
            self.executed = False
            return None

        if self.executed:
            return None

        Events.post_event(
            self.name,
            (ValueNoChangeEvent if self.__watch_only_changes else ValueNoUpdateEvent)
            (self.name, item.state, int(duration.total_seconds()))
        )
        self.executed = True

    def cancel(self):
        self.is_canceled = True



import asyncio

import HABApp.config
import HABApp.core
import HABApp.rule_manager
import HABApp.util

from .folder_watcher import FolderWatcher
from .shutdown_helper import ShutdownHelper


class Runtime:
    def __init__(self, config_folder):

        self.loop = asyncio.get_event_loop()

        self.shutdown = ShutdownHelper()

        self.folder_watcher = FolderWatcher()
        self.folder_watcher.start(self.shutdown)

        self.config = HABApp.config.Config(self, config_folder=config_folder)

        # OpenHAB
        self.openhab_connection = HABApp.openhab.OpenhabConnection(self)

        # MQTT
        self.mqtt_connection = HABApp.mqtt.MqttConnection(self)
        self.mqtt_connection.connect()

        self.rule_manager = HABApp.rule_manager.RuleManager(self)
        self.rule_params = HABApp.rule_manager.RuleParameters(self)


        # Shutdown workers
        self.shutdown.register_func(HABApp.core.WrappedFunction._WORKERS.shutdown)

    @HABApp.util.PrintException
    def get_async(self):
        return asyncio.gather(
            self.openhab_connection.start(),
            self.rule_manager.get_async(),
        )



import asyncio
import logging
import traceback
import typing

import aiohttp
import ujson
from aiohttp.client import ClientResponse
from aiohttp_sse_client import client as sse_client

import HABApp
import HABApp.core
import HABApp.openhab.events

log = logging.getLogger('HABApp.openhab.connection')
log_events = logging.getLogger('HABApp.Events.openhab')


class OpenhabDisconnectedError(Exception):
    pass


class OpenhabNotReadyYet(Exception):
    pass


class HttpConnectionEventHandler:
    def on_connected(self):
        raise NotImplementedError()

    def on_disconnected(self):
        raise NotImplementedError()

    def on_sse_event(self, event: dict):
        raise NotImplementedError()


class HttpConnection:

    def __init__(self, event_handler, config):
        assert isinstance(event_handler, HttpConnectionEventHandler)
        self.event_handler: HttpConnectionEventHandler = event_handler

        assert isinstance(config, HABApp.config.Config)
        self.config: HABApp.config.Config = config

        self.is_online = False
        self.is_read_only = False

        self.__session: aiohttp.ClientSession = None

        self.__host: str = ''
        self.__port: str = ''

        self.__wait = 0

        self.async_try_uuid: asyncio.Future = None

        # automatically update config
        self.__update_config_general()
        self.config.openhab.general.subscribe_for_changes(self.__update_config_general)

    def __update_config_general(self):
        self.is_read_only = self.config.openhab.general.listen_only

    def __get_openhab_url(self, url: str, *args, **kwargs) -> str:
        assert not url.startswith('/')
        url = url.format(*args, **kwargs)
        return f'http://{self.__host:s}:{self.__port:d}/{url:s}'

    def __set_offline(self):

        if not self.is_online:
            return None
        self.is_online = False

        self.__wait = 5
        self.event_handler.on_disconnected()

        # Try reconnect
        if not self.async_try_uuid.done():
            self.async_try_uuid.cancel()
        self.async_try_uuid = asyncio.run_coroutine_threadsafe(self._try_uuid(), asyncio.get_event_loop())

    def _is_disconnect_exception(self, e) -> bool:
        if not isinstance(e, (aiohttp.ClientPayloadError, aiohttp.ClientConnectorError)):
            return False

        self.__set_offline()
        return True

    async def _check_http_response(self, future, additional_info = "") -> typing.Optional[ ClientResponse]:
        try:
            resp = await future
        except Exception as e:
            is_disconnect = self._is_disconnect_exception(e)
            log.log(logging.WARNING if is_disconnect else logging.ERROR, f'{e}')
            if is_disconnect:
                raise OpenhabDisconnectedError()
            else:
                return None

        # Server Errors if openhab is not ready yet
        if resp.status >= 500:
            self.__set_offline()
            raise OpenhabNotReadyYet()

        # Something went wrong - log error message
        if resp.status >= 300:
            # Log Error Message
            additional_info = f' ({additional_info})' if additional_info else ""
            log.warning(f'Status {resp.status} for {resp.request_info.method} {resp.request_info.url}{additional_info}')
            for line in str(resp).splitlines():
                log.warning(line)

        return resp

    def cancel_connect(self):
        if not self.async_try_uuid:
            return None

        if not self.async_try_uuid.done():
            self.async_try_uuid.cancel()

    async def try_connect(self):

        self.cancel_connect()

        # If we are already connected properly disconnect
        if self.__session is not None:
            await self.__session.close()
            self.__session = None

        self.__host: str = self.config.openhab.connection.host
        self.__port: str = self.config.openhab.connection.port

        # do not run without host
        if self.__host == '':
            return None

        auth = None
        if self.config.openhab.connection.user or self.config.openhab.connection.password:
            auth = aiohttp.BasicAuth(
                self.config.openhab.connection.user,
                self.config.openhab.connection.password
            )

        self.__session = aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(total=99999999999999999),
            json_serialize=ujson.dumps,
            auth=auth
        )

        self.async_try_uuid = asyncio.ensure_future(self._try_uuid())

    async def _try_uuid(self):

        # sleep before reconnect
        await asyncio.sleep(self.__wait)
        self.__wait *= 2
        self.__wait = min(self.__wait, 180)
        self.__wait = max(5, 0)

        log.debug('Trying to connect to OpenHAB ...')
        try:
            uuid = await self.async_get_uuid()
        except OpenhabDisconnectedError or OpenhabNotReadyYet:
            self.async_try_uuid = asyncio.ensure_future(self._try_uuid())
            log.info(f'... offline!')
            return None

        log.info( f'Connected to OpenHAB instance {uuid}')
        self.is_online = True
        self.event_handler.on_connected()
        return None

    async def async_process_sse_events(self):
        try:
            # cache so we don't have to look up every event
            call = self.event_handler.on_sse_event

            async with sse_client.EventSource(
                    self.__get_openhab_url("rest/events?topics=smarthome/items/"),
                    session=self.__session
            ) as event_source:
                async for event in event_source:
                    event = ujson.loads(event.data)

                    # Log sse event
                    if log_events.isEnabledFor(logging.DEBUG):
                        log_events._log(logging.DEBUG, event, [])

                    # process
                    call(event)

        except asyncio.CancelledError:
            pass

        except Exception as e:
            disconnect = self._is_disconnect_exception(e)
            lvl = logging.WARNING if disconnect else logging.ERROR
            log.log(lvl, f'SSE request Error: {e}')
            for l in traceback.format_exc().splitlines():
                log.log(lvl, l)
            if not disconnect:
                raise

    async def async_post_update(self, item, state):

        if self.config.openhab.general.listen_only:
            return False

        fut = self.__session.put(self.__get_openhab_url('rest/items/{item:s}/state', item=item), data=state)
        asyncio.ensure_future(self._check_http_response(fut))

    async def async_send_command(self, item, state):

        if self.config.openhab.general.listen_only:
            return False

        fut = self.__session.post(self.__get_openhab_url('rest/items/{item:s}', item=item), data=state)
        asyncio.ensure_future(self._check_http_response(fut))

    async def async_remove_item(self, item_name) -> bool:

        if self.config.openhab.general.listen_only:
            return False

        fut = self.__session.delete(self.__get_openhab_url('rest/items/{:s}', item_name))
        ret = await self._check_http_response(fut)
        return ret.status < 300

    async def async_get_uuid(self) -> str:
        fut = self.__session.get(self.__get_openhab_url('rest/uuid'))
        resp = await self._check_http_response(fut)
        return (await resp.text()) if resp else resp

    async def async_get_items(self) -> typing.Optional[list]:
        fut = self.__session.get(
            self.__get_openhab_url('rest/items'),
            json={'recursive': 'false', 'fields': 'state,type,name,editable'}
        )

        resp = await self._check_http_response(fut)
        return ujson.loads(await resp.text())

    async def async_create_item(self, item_type, item_name, label="", category="", tags=[], groups=[]) -> bool:

        if self.config.openhab.general.listen_only:
            return False


        payload = {'type': item_type, 'name': item_name}
        if label:
            payload['label'] = label
        if category:
            payload['category'] = category
        if tags:
            payload['tags'] = tags
        if groups:
            payload['groupnames'] = groups

        fut = self.__session.put(self.__get_openhab_url('rest/items/{:s}', item_name), json=payload)
        ret = await self._check_http_response(fut, payload)
        return ret.status < 300



import logging
import typing
import itertools

from HABApp.util import PrintException
from .worker import WrappedFunction

event_log = logging.getLogger('HABApp.Events')
habapp_log = logging.getLogger('HABApp')


class AllEvents:
    pass


class ValueUpdateEvent:
    def __init__(self, name = None, value = None):
        self.name: str = name
        self.value = value

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class ValueChangeEvent:
    def __init__(self, name = None, value = None, old_value = None):
        self.name: str = name
        self.value = value
        self.old_value = old_value

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}, old_value: {self.old_value}>'


class ValueNoChangeEvent:
    def __init__(self, name = None, value = None, seconds = None):
        self.name: str = name
        self.value = value
        self.seconds: int = seconds

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class ValueNoUpdateEvent:
    def __init__(self, name = None, value = None, seconds = None):
        self.name: str = name
        self.value = value
        self.seconds: int = seconds

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class EventListener:
    def __init__(self, name, callback, event_type=AllEvents):
        assert isinstance(name, str) or name is None, type(name)
        assert isinstance(callback, WrappedFunction)

        self.name: str = name
        self.func = callback

        self.event_filter = event_type

    def notify_listeners(self, event):

        if self.event_filter is AllEvents or isinstance(event, self.event_filter):
            self.func.submit(event)
            return None

        # Make it possible to specify multiple classes
        if isinstance(self.event_filter, list) or isinstance(self.event_filter, set):
            for cls in self.event_filter:
                if isinstance(event, cls):
                    self.func.submit(event)
                    return None


class EventBus:
    def __init__(self):

        from . import Items
        self.__items = Items

        self.__event_listener: typing.Dict[str, typing.List[EventListener]] = {}
        self.__event_listener_all: typing.List[EventListener] = []

    @PrintException
    def post_event(self, name, event):

        event_log.info(event)

        # Update Item Registry BEFORE doing the callbacks
        if isinstance(event, ValueUpdateEvent):
            self.__items.set_state(event.name, event.value)

        # Notify all listeners
        for listener in itertools.chain(self.__event_listener.get(name, []), self.__event_listener_all):
            listener.notify_listeners(event)

        return None

    def remove_listener(self, listener : EventListener):
        assert isinstance(listener, EventListener)
        add_to_all = listener.name is None

        item_listeners = self.__event_listener.get(listener.name, []) if not add_to_all else self.__event_listener_all
        if listener not in item_listeners:
            return None
        item_listeners.remove(listener)

        if add_to_all:
            habapp_log.debug(f'Removed event listener for all names (type {listener.event_filter})')
        else:
            habapp_log.debug(f'Removed event listener for "{listener.name}" (type {listener.event_filter})')

    def add_listener(self, listener : EventListener):
        assert isinstance(listener, EventListener)
        add_to_all = listener.name is None

        # don't add the same listener twice
        item_listeners = self.__event_listener.get(listener.name, []) if not add_to_all else self.__event_listener_all
        if listener in item_listeners:
            return None

        item_listeners.append( listener)
        if add_to_all:
            habapp_log.debug(f'Added Event listener for all names (type {listener.event_filter})')
        else:
            self.__event_listener[listener.name] = item_listeners
            habapp_log.debug(f'Added Event listener for "{listener.name}" (type {listener.event_filter})')
        return None



import HABApp.openhab.items

from .oh_interface import OpenhabInterface
from .oh_connection import OpenhabConnection
from .http_connection import HttpConnection
from .items.map_items import map_items




import argparse
import asyncio
import logging
import os
import signal
import sys
import traceback
import typing
from pathlib import Path

import HABApp

parser = argparse.ArgumentParser(description='Start HABApp')
parser.add_argument(
    '-c',
    '--config',
    help='Path to configuration folder (where the config.yml is located)"',
    default=None
)
args = parser.parse_args()
if args.config is not None:
    args.config = Path(args.config).resolve()


def find_config_folder() -> Path:
    check_path = [
        Path(os.getcwd()) / 'HABApp',                          # current working dir
        Path(os.environ.get('VIRTUAL_ENV', '')) / 'HABApp',    # Virtual env dir
        Path.home() / 'HABApp',                                #
    ]
    if args.config is not None:
        check_path = [args.config]

    for p in check_path:
        p = p.resolve()
        if not p.is_dir():
            continue

        f = p / 'config.yml'
        if f.is_file():
            return p

    # we have specified a folder, if the config does not exist we will create it
    if args.config is not None and args.config.is_dir():
        return args.config

    # we have nothing found and nothing specified -> exit
    print('Config file "config.yml" not found!')
    print('Checked folders:\n - ' + '\n - '.join(str(k) for k in check_path if str(k) != 'HABApp'))
    print('Please create file or specify a folder with the "-c" arg switch.')
    sys.exit(1)



def main() -> typing.Union[int, str]:
    loop = None
    log = logging.getLogger('HABApp')

    try:
        loop = asyncio.get_event_loop()
        loop.set_debug(True)

        app = HABApp.Runtime(find_config_folder())

        def shutdown_handler(sig, frame):
            print('Shutting down ...')
            app.shutdown.request_shutdown()

        # register shutdown helper
        signal.signal(signal.SIGINT, shutdown_handler)
        signal.signal(signal.SIGTERM, shutdown_handler)

        # start workers
        try:
            loop.run_until_complete(app.get_async())
        except asyncio.CancelledError:
            pass
    except Exception as e:
        for line in traceback.format_exc().splitlines():
            log.error(line)
            print(e)
        return str(e)
    finally:
        loop.close()
    return 0


if __name__ == "__main__":
    sys.exit(main())



from .worker import WrappedFunction

from .events import EventBus as __EventBus
from .events import EventListener, ValueUpdateEvent, ValueChangeEvent, ValueNoChangeEvent, ValueNoUpdateEvent, AllEvents
from .items import Items as __Items
from .items import Item

Items = __Items()
Events = __EventBus()



from voluptuous import All, Invalid, Length

from .configentry import ConfigEntry, ConfigEntryContainer


def MqttTopicValidator(msg=None):
    def f(v):
        if isinstance(v, str):
            return [(v, 0)]

        ret = []
        for i, val in enumerate(v):
            qos = 0
            if i < len(v) - 1:
                qos = v[i + 1]

            if not isinstance(val, str) and not isinstance(val, int):
                raise Invalid(msg or (f"Topics must consist of int and string!"))

            if not isinstance(val, str):
                continue

            if isinstance(qos, int):
                if qos not in [0, 1, 2]:
                    raise Invalid(msg or (f"QoS must be 0, 1 or 2"))
            else:
                qos = None

            ret.append((val, qos))
        return ret

    return f


class MqttConnection(ConfigEntry):
    def __init__(self):
        super().__init__()
        self._entry_name = 'connection'
        self.client_id = 'HABApp'
        self.host = ''
        self.port = 8883
        self.user = ''
        self.password = ''
        self.tls = True
        self.tls_insecure = False

        self._entry_validators['client_id'] = All(str, Length(min=1))

        self._entry_kwargs['password'] = {'default': ''}
        self._entry_kwargs['tls_insecure'] = {'default': False}


class Subscribe(ConfigEntry):
    def __init__(self):
        super().__init__()
        self.qos = 0
        self.topics = ['#', 0]

        self._entry_validators['topics'] = MqttTopicValidator()
        self._entry_kwargs['topics'] = {'default': [('#', 0)]}
        self._entry_kwargs['default_qos'] = {'default': 0}


class Publish(ConfigEntry):
    def __init__(self):
        super().__init__()
        self.qos = 0
        self.retain = False


class mqtt(ConfigEntry):
    def __init__(self):
        super().__init__()
        self.client_id = ''
        self.tls = True
        self.tls_insecure = False

        self._entry_kwargs['tls_insecure'] = {'default': False}


class Mqtt(ConfigEntryContainer):
    def __init__(self):
        self.connection = MqttConnection()
        self.subscribe = Subscribe()
        self.publish = Publish()



import ujson


class BaseItemEvent:
    def __init__(self, _in_dict: dict):
        super().__init__()
        assert isinstance(_in_dict, dict), type(_in_dict)
        self._topic = _in_dict['topic']
        self._type = _in_dict['type']
        self._payload = ujson.loads(_in_dict['payload'])

    def __repr__(self):
        return f'<{self.__class__.__name__} _topic: {self._topic} _type: {self._type}, _payload: {self._payload}>'



from HABApp.core.items import Item


class ContactItem(Item):
    OPEN = 'OPEN'
    CLOSED = 'CLOSED'

    @classmethod
    def from_str(self, name, value):
        item = ContactItem(name=name)
        item.set_state(value)
        return item

    def set_state(self, new_state):
        if new_state is not None and new_state != ContactItem.OPEN and new_state != ContactItem.CLOSED:
            raise ValueError(f'Invalid value for ContactItem: {new_state}')
        super().set_state(new_state)

    def is_open(self):
        return True if self.state == ContactItem.OPEN else False

    def is_closed(self):
        return True if self.state == ContactItem.CLOSED else False

    def __str__(self):
        return self.state

    def __eq__(self, other):
        if isinstance(other, ContactItem):
            return self.state == other.state
        elif isinstance(other, str):
            return self.state == other
        elif isinstance(other, int):
            if other and self.is_open():
                return True
            if not other and self.is_closed():
                return True
            return False

        return NotImplemented

    def __ne__(self, other):
        res = self.__eq__(other)
        if res is NotImplemented:
            return res
        return not res



import asyncio
import datetime
import logging

import HABApp
import HABApp.classes
import HABApp.core
import HABApp.openhab.events
from HABApp.config import Openhab as OpenhabConfig
from HABApp.util import PrintException
from .http_connection import HttpConnection

log = logging.getLogger('HABApp.openhab.Connection')


class OpenhabInterface:
    def __init__(self, connection, openhab_config):
        assert isinstance(openhab_config, OpenhabConfig)

        self.__config: OpenhabConfig = openhab_config
        self.__loop = asyncio.get_event_loop()
        self.__connection: HttpConnection = connection

    def __convert_to_oh_type(self, _in):
        if isinstance(_in, datetime.datetime):
            return _in.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + self.__config.general.timezone
        elif isinstance(_in, HABApp.core.Item):
            return str(_in.state)
        elif isinstance(_in, HABApp.classes.Color):
            return f'{_in.hue:.1f},{_in.saturation:.1f},{_in.value:.1f}'

        return str(_in)

    @PrintException
    def post_update(self, item_name: str, state):
        if not self.__connection.is_online or self.__connection.is_read_only:
            return None

        asyncio.run_coroutine_threadsafe(
            self.__connection.async_post_update(item_name, self.__convert_to_oh_type(state)),
            self.__loop
        )

    @PrintException
    def send_command(self, item_name: str, state):
        if not self.__connection.is_online or self.__connection.is_read_only:
            return None

        asyncio.run_coroutine_threadsafe(
            self.__connection.async_send_command(item_name, self.__convert_to_oh_type(state)),
            self.__loop
        )


    @PrintException
    def create_item(self, item_type: str, item_name: str, label="", category="", tags=[], groups=[]):
        if not self.__connection.is_online or self.__connection.is_read_only:
            return None

        assert isinstance(item_type, str), type(item_type)
        item_type = item_type.title()
        assert item_type in ['String', 'Number', 'Switch', 'Contact', 'Color', 'Contact'], item_type
        assert isinstance(item_name, str), type(item_name)
        assert isinstance(label, str), type(label)
        assert isinstance(category, str), type(category)
        assert isinstance(tags, list), type(tags)
        assert isinstance(groups, list), type(groups)

        fut = asyncio.run_coroutine_threadsafe(
            self.__connection.async_create_item(item_type, item_name, label, category, tags, groups),
            self.__loop
        )
        return fut.result()

    @PrintException
    def remove_item(self, item_name: str):
        if not self.__connection.is_online or self.__connection.is_read_only:
            return None

        fut = asyncio.run_coroutine_threadsafe(
            self.__connection.async_remove_item(item_name),
            self.__loop
        )
        return fut.result()



import collections
import statistics
import time


class Statistics:
    def __init__(self, max_age=None, max_samples=None):
        """
        Keep stats
        :param max_age:     Maximum age in seconds
        :param max_samples: Maximum amount of samples
        """

        if max_age is None and max_samples is None:
            raise ValueError('Please specify max age or max samples!')

        self._max_age = max_age

        self.timestamps = collections.deque(maxlen=max_samples)
        self.states = collections.deque(maxlen=max_samples)

        self.sum: float = None
        self.min: float = None
        self.max: float = None

        self.mean: float = None
        self.median: float = None

        self.last_value: float = None
        self.last_change: float = None

    def _remove_old(self):
        if self._max_age is None:
            return None

        # remove too old entries
        now = time.time()
        while self.timestamps and (now - self.timestamps[0]) > self._max_age:
            self.timestamps.popleft()
            self.states.popleft()

    def update(self):
        self._remove_old()

        __len = len(self.states)

        if not __len:
            self.sum = None
            self.min = None
            self.max = None

            self.mean = None
            self.median = None
        else:
            self.sum = sum(self.states)
            self.min = min(self.states)
            self.max = max(self.states)

            self.mean = statistics.mean(self.states)
            self.median = statistics.median(self.states)

        if __len >= 2:
            self.last_change = self.states[-1] - self.states[-2]
        else:
            self.last_change = None

    def add_value(self, value):
        assert isinstance(value, (int, float)), type(value)

        self.last_value = value

        self.timestamps.append(time.time())
        self.states.append(value)

        self.update()

    def __repr__(self):
        return f'<Statistics sum: {self.sum:.1f}, min: {self.min:.2f}, max: {self.max:.2f}, ' \
               f'mean: {self.mean:.2f}, median: {self.median:.2f}>'



import collections
import logging
import runpy
import traceback
import typing
from pathlib import Path

if typing.TYPE_CHECKING:
    import HABApp

log = logging.getLogger(f'HABApp.Rules')


class RuleFile:
    def __init__(self, rule_manager, path: Path):
        from .rule_manager import RuleManager

        assert isinstance(rule_manager, RuleManager)
        self.rule_manager = rule_manager

        self.path = path

        self.rules = {}

        self.class_ctr = collections.defaultdict(lambda : 1)

    def suggest_rule_name(self, obj) -> str:

        # if there is already a name set we make no suggestion
        if getattr(obj, 'rule_name', '') != '':
            return obj.rule_name.replace('ü', 'ue').replace('ö', 'oe').replace('ä', 'ae')

        # create unique name
        name = f'{str(type(obj))[19:-2]:s}'
        found = self.class_ctr[name]
        self.class_ctr[name] += 1

        return f'{name:s}.{found:d}' if found > 1 else f'{name:s}'

    def iterrules(self):
        for rule in self.rules.values():
            yield rule

    def check_all_rules(self):
        for rule in self.rules.values():  # type: HABApp.Rule
            rule._check_rule()

    def unload(self):

        # If we don't have any rules we can not unload
        if not self.rules:
            return None

        # unload all registered callbacks
        for rule in self.rules.values():  # type: HABApp.Rule
            rule._cleanup()

        log.debug(f'File {self.path} successfully unloaded!')
        return None

    def load(self):

        created_rules = []
        try:
            runpy.run_path(self.path, init_globals={
                '__HABAPP__RUNTIME__': self.rule_manager.runtime,
                '__HABAPP__RULE_FILE__': self,
                '__HABAPP__RULES' : created_rules
            })
        except Exception as e:
            log.error(f"Could not load {self.path}: {e}!")
            for l in traceback.format_exc().splitlines()[-5:]:
                log.error(l)

        len_found = len(created_rules)
        if not len_found:
            log.warning(f'Found no instances of HABApp.Rule in {str(self.path)}')
        else:
            for rule in created_rules:
                # ensure that we have a rule name
                rule.rule_name = self.suggest_rule_name(rule)

                # rule name must be unique for every file
                if rule.rule_name in self.rules:
                    raise ValueError(f'Rule name must be unique!\n"{rule.rule_name}" is already used!')

                self.rules[rule.rule_name] = rule
                log.info(f'Added rule "{rule.rule_name}" from {self.path.name}')

        return None



import pathlib

from voluptuous import Required, Coerce, Optional


class ConfigEntry:
    def __init__(self):
        self._entry_is_required = True
        self._entry_name = self.__class__.__name__.lower()
        self._entry_kwargs = {}
        self._entry_validators = {}
        self._entry_notify_on_change = []

    def subscribe_for_changes(self, callback):
        assert callback not in self._entry_notify_on_change
        self._entry_notify_on_change.append(callback)

    def iter_entry(self):
        for name, value in self.__dict__.items():
            assert name.islower(), f'variable name must be lower case! "{name}"'
            if name.startswith('_entry_'):
                continue
            yield name, value

    def update_schema(self, _dict):

        val = {}
        for name, value in self.iter_entry():

            # datatype
            if name in self._entry_validators:
                _type = self._entry_validators[name]
            else:
                _type = type(value)
                if _type is int or _type is float:
                    _type = Coerce(_type)

                # we do not load Path objects, we load the strings
                if isinstance(value, pathlib.Path):
                    _type = str

            # name
            __name = {'schema' : name}
            __name.update(self._entry_kwargs.get(name, {}))
            val[Required(**__name)] = _type

        _dict[ Required(self._entry_name) if self._entry_is_required else Optional(self._entry_name)] = val
        return _dict

    def insert_data(self, _dict):
        _insert = {}
        for name, value in self.iter_entry():
            _insert[name] = value
        _dict[self._entry_name] = _insert

    def load_data(self, _dict):
        if self._entry_name not in _dict:
            return None

        notify = False
        _dict = _dict[self._entry_name]
        for name, value in self.iter_entry():
            cur = getattr(self, name)
            new = _dict[name]
            if cur != new:
                notify = True
            setattr(self, name, new)

        if notify:
            for cb in self._entry_notify_on_change:
                cb()


class ConfigEntryContainer:

    def iter_entriy(self):
        for name, value in self.__dict__.items():
            if not isinstance(value, ConfigEntry):
                continue
            yield name, value

    def insert_data(self, _dict):
        _insert = {}
        for name, value in self.iter_entriy():
            value.insert_data(_insert)
        _dict[self.__class__.__name__.lower()] = _insert

    def load_data(self, _dict):
        _load = _dict[self.__class__.__name__.lower()]
        for name, value in self.iter_entriy():
            value.load_data(_load)

    def update_schema(self, _dict):
        schema = {}
        for name, value in self.iter_entriy():
            value.update_schema(schema)
        _dict[Required(self.__class__.__name__.lower())] = schema



import datetime

from HABApp.classes import Color
from HABApp.core.items import Item

from . import SwitchItem, ContactItem


def map_items(name, openhab_type : str, openhab_value : str):
    assert isinstance(openhab_type, str), type(openhab_type)
    assert isinstance(openhab_value, str), type(openhab_value)

    value = openhab_value
    if openhab_value == 'NULL' or openhab_value == 'UNDEF':
        value = None

    # Specific classes
    if openhab_type == "Switch":
        return SwitchItem.from_str(name, value)
    if openhab_type == "Contact":
        return ContactItem.from_str(name, value)

    item = Item(name)
    if value is None:
        item.set_state(value)
        return item

    if openhab_type == "Number":
        try:
            value = int(openhab_value)
        except ValueError:
            value = float(openhab_value)

    elif openhab_type == "Decimal":
        value = float(openhab_value)

    elif openhab_type == "Color":
        value = Color(*[float(k) for k in openhab_value.split(',')])

    elif openhab_type == "DateTime":
        value = datetime.datetime.strptime(openhab_value.replace('+', '000+'), '%Y-%m-%dT%H:%M:%S.%f%z')

    item.set_state(value)
    return item



import functools
import traceback


def PrintException( func):

    @functools.wraps(func)
    def f(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            print("{}\n{}".format( e, traceback.format_exc()))
            raise
    return f



import colorsys
import typing


class Color:
    HUE_FACTOR = 360
    PERCENT_FACTOR = 100

    def __init__(self, h=0.0, s=0.0, v=0.0):
        self.hue = min(max(0, h), Color.HUE_FACTOR)
        self.saturation = min(max(0, s), Color.PERCENT_FACTOR)
        self.value = min(max(0, v), Color.PERCENT_FACTOR)

    def get_rgb(self, max_rgb_value=255) -> typing.Tuple[int, int, int]:
        r, g, b = colorsys.hsv_to_rgb(
            self.hue / Color.HUE_FACTOR,
            self.saturation / Color.PERCENT_FACTOR,
            self.value / Color.PERCENT_FACTOR
        )
        return int(r * max_rgb_value), int(g * max_rgb_value), int(b * max_rgb_value)

    def set_rgb(self, r, g, b, max_rgb_value=255):
        h, s, v = colorsys.rgb_to_hsv(r / max_rgb_value, g / max_rgb_value, b / max_rgb_value)
        self.hue = h * Color.HUE_FACTOR
        self.saturation = s * Color.PERCENT_FACTOR
        self.value = v * Color.PERCENT_FACTOR
        return self

    def __repr__(self):
        return f'<Color Hue: {self.hue}°, Saturation: {self.saturation}%, Value: {self.value}%>'



import datetime
import logging
import typing

if typing.TYPE_CHECKING:
    pass

log = logging.getLogger('HABApp.Items')


class Item:
    def __init__(self, name: str):
        assert isinstance(name, str), type(name)

        self.name = name
        self.state = None

        _now = datetime.datetime.now()
        self.last_change: datetime.datetime = _now
        self.last_update: datetime.datetime = _now

    def set_state(self, new_state):
        _now = datetime.datetime.now()
        if self.state != new_state:
            self.last_change = _now
        self.last_update = _now
        self.state = new_state

    def __repr__(self):
        ret = f'<{self.__class__.__name__}'
        for k in ['name', 'state', 'last_change', 'last_update']:
            ret += f' {k}: {getattr(self, k)}'

        return ret + '>'


class Items:
    def __init__(self):
        self.items = {}  # type: typing.Dict[str,Item]

    def item_exists(self, name) -> bool:
        return name in self.items

    def get_item(self, name) -> Item:
        return self.items[name]

    def get_items(self) -> typing.List[Item]:
        return list(self.items.values())

    def get_item_names(self) -> typing.List[str]:
        return list(self.items.keys())

    def set_state(self, name, new_state):
        try:
            self.items[name].set_state(new_state)
        except KeyError:
            item = Item(name)
            item.set_state(new_state)
            self.items[name] = item

    def set_item(self, item):
        assert isinstance(item, Item), type(item)
        self.items[item.name] = item

    def pop_item(self, name) -> Item:
        return self.items.pop(name)



import asyncio
import datetime
import logging
import math
import threading
import time
import traceback
import typing
from pathlib import Path

import HABApp
from HABApp.util import PrintException
from .rule_file import RuleFile
from HABApp.runtime import FileEventTarget

log = logging.getLogger('HABApp.Rules')


class RuleManager(FileEventTarget):

    def __init__(self, parent):
        assert isinstance(parent, HABApp.Runtime)
        self.runtime = parent

        self.files: typing.Dict[str, RuleFile] = {}

        # serialize loading
        self.__lock = threading.Lock()

        # if we load immediately we don't have the items from openhab in itemcache
        def delayed_load():
            time.sleep(5)
            for f in self.runtime.config.directories.rules.glob('**/*.py'):
                if f.name.endswith('.py'):
                    time.sleep(0.5)
                    HABApp.core.WrappedFunction(self.add_file, logger=log).submit( f)

        HABApp.core.WrappedFunction(delayed_load, logger=log, warn_too_long=False).submit()

        # folder watcher
        self.runtime.folder_watcher.watch_folder(
            folder=self.runtime.config.directories.rules,
            file_ending='.py',
            event_target=self,
            worker_factory=lambda x : HABApp.core.WrappedFunction(x, logger=log).submit,
            watch_subfolders=True
        )

        self.__process_last_sec = 60

    @PrintException
    async def process_scheduled_events(self):

        while not self.runtime.shutdown.requested:

            now = datetime.datetime.now()

            # process only once per second
            if now.second == self.__process_last_sec:
                await asyncio.sleep(0.1)
                continue
            self.__process_last_sec = now.second

            for file in self.files.values():
                assert isinstance(file, RuleFile), type(file)
                for rule in file.iterrules():
                    rule._process_events(now)

            # sleep longer, try to sleep until the next full second
            end = datetime.datetime.now()
            if end.second == self.__process_last_sec:
                frac, whole = math.modf(time.time())
                sleep_time = 1 - frac + 0.005   # prevent rounding error and add a little bit of security
                await asyncio.sleep( sleep_time)


    @PrintException
    def get_async(self):
        return asyncio.gather(self.process_scheduled_events())

    @PrintException
    def get_rule(self, rule_name):
        found = []
        for file in self.files.values():
            if rule_name in file.rules:
                found.append(file.rules[rule_name])
        if not found:
            raise KeyError(f'No Rule with name "{rule_name}" found!')
        return found if len(found) > 1 else found[0]

    @PrintException
    def reload_file(self, path: Path):
        self.remove_file(path)
        self.add_file(path)

    @PrintException
    def remove_file(self, path: Path):
        try:
            with self.__lock:
                path_str = str(path)
                log.debug(f'Removing file: {path}')
                if path_str in self.files:
                    self.files.pop(path_str).unload()
        except Exception:
            log.error(f"Could not remove {path}!")
            for l in traceback.format_exc().splitlines():
                log.error(l)
            return None

    @PrintException
    def add_file(self, path : Path):

        try:
            # serialize loading
            with self.__lock:
                path_str = str(path)

                log.debug(f'Loading file: {path}')
                self.files[path_str] = file = RuleFile(self, path)
                file.load()
        except Exception:
            log.error(f"Could not (fully) load {path}!")
            for l in traceback.format_exc().splitlines():
                log.error(l)
            return None

        log.debug(f'File {path_str} successfully loaded!')

        # Do simple checks which prevent errors
        file.check_all_rules()



from .scheduled_cb import ScheduledCallback, TYPING_DATE_TIME
from .reoccurring_cb import ReoccurringScheduledCallback, DayOfWeekScheduledCallback, \
    WeekendScheduledCallback, WorkdayScheduledCallback, TYPING_TIME



from pathlib import Path
import setuptools

# Load version number
version = {}
with open("HABApp/__version__.py") as fp:
    exec(fp.read(), version)
assert version
assert version['__VERSION__']
__VERSION__ = version['__VERSION__']
print(f'Version: {__VERSION__}')
print('')

# don't load file for tox-builds
readme = Path(__file__).with_name('readme.md')
long_description = ''
if readme.is_file():
    with open(readme, "r", encoding='utf-8') as fh:
        long_description = fh.read()

setuptools.setup(
    name="HABApp",
    version=__VERSION__,
    author="spaceman_spiff",
    # author_email="",
    description="Easy automation with openHAB and/or MQTT. Create home automation rules in python.",
    keywords='openHAB,habapp,mqtt,home,automation',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/spacemanspiff2007/HABApp",
    packages=setuptools.find_packages(exclude=['tests*']),
    install_requires=[
        'aiohttp',
        'aiohttp-sse-client',
        'ruamel.yaml',
        'paho-mqtt',
        'ujson',
        'voluptuous',
        'watchdog',
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'habapp = HABApp.__main__:main'
        ]
    }
)



import datetime
import random
import sys
import typing
import logging

import HABApp
import HABApp.core
import HABApp.openhab
import HABApp.rule_manager
import HABApp.util
import HABApp.classes
from .watched_item import WatchedItem
from .rule_parameter import RuleParameter
from .scheduler import ReoccurringScheduledCallback, ScheduledCallback, WorkdayScheduledCallback, \
    WeekendScheduledCallback, DayOfWeekScheduledCallback, TYPING_DATE_TIME, TYPING_TIME

log = logging.getLogger('HABApp.Rule')


class Rule:
    def __init__(self):

        # get the variables from the caller
        depth = 1
        while True:
            __vars = sys._getframe(depth).f_globals
            depth += 1
            if '__HABAPP__RUNTIME__' in __vars:
                __runtime__ = __vars['__HABAPP__RUNTIME__']
                __rule_file__ = __vars['__HABAPP__RULE_FILE__']
                break

        # this is a list which contains all rules of this file
        __vars['__HABAPP__RULES'].append(self)

        # assert isinstance(__runtime__, HABApp.Runtime)
        self.__runtime: HABApp.Runtime = __runtime__

        # assert isinstance(__rule_file__, HABApp.rule_manager.RuleFile)
        self.__rule_file: HABApp.rule_manager.RuleFile = __rule_file__

        self.__event_listener: typing.List[HABApp.core.EventListener] = []
        self.__future_events: typing.List[ScheduledCallback] = []
        self.__watched_items: typing.List[WatchedItem] = []

        # suggest a rule name if it is not
        self.rule_name: str = self.__rule_file.suggest_rule_name(self)

        # interfaces
        self.mqtt = self.__runtime.mqtt_connection.interface
        self.oh: HABApp.openhab.OpenhabInterface = self.__runtime.openhab_connection.interface
        self.openhab: HABApp.openhab.OpenhabInterface = self.oh

    def item_exists(self, name: str) -> bool:
        """
        Checks whether an item exists
        :param name: Name of the item
        :return: True or False
        """
        assert isinstance(name, str), type(name)
        return HABApp.core.Items.item_exists(name)

    def get_item_state(self, name: str, default=None):
        """
        Return the state of the item.
        :param name:
        :param default: If the item does not exist or is None this value will be returned (has to be != None)
        :return: state of the specified item
        """
        if default is None:
            return HABApp.core.Items.get_item(name).state

        try:
            state = HABApp.core.Items.get_item(name).state
        except KeyError:
            return default

        if state is None:
            return default
        return state

    def set_item_state(self, name: str, value):
        assert isinstance(name, str)

        try:
            old_state = HABApp.core.Items.get_item(name).state
        except KeyError:
            old_state = None

        self.post_event(name, HABApp.core.ValueUpdateEvent(name=name, value=value))
        if old_state != value:
            self.post_event(name, HABApp.core.ValueChangeEvent(name=name, value=value, old_value=old_state))
        return None

    def item_watch(self, name: str, seconds_constant: int, watch_only_changes=True) -> WatchedItem:
        assert isinstance(name, str)
        assert isinstance(seconds_constant, int)
        assert isinstance(watch_only_changes, bool)

        item = WatchedItem(
            name=name,
            constant_time=seconds_constant,
            watch_only_changes=watch_only_changes
        )
        self.__watched_items.append(item)
        return item

    def item_watch_and_listen(self, name: str, seconds_constant: int, callback,
                              watch_only_changes = True) -> typing.Tuple[WatchedItem, HABApp.core.EventListener]:

        watched_item = self.item_watch(name, seconds_constant, watch_only_changes)
        event_listener = self.listen_event(
            name,
            callback,
            HABApp.core.ValueNoChangeEvent if watch_only_changes else HABApp.core.ValueNoUpdateEvent
        )
        return watched_item, event_listener

    def get_item(self, name: str) -> HABApp.core.Item:
        return HABApp.core.Items.get_item(name)

    def post_event(self, name, event):
        """
        Post an Event to the Event Bus
        :param name: name to post event to
        :param event: Event class to be used (must be class instance)
        :return:
        """
        assert isinstance(name, str), type(name)
        return HABApp.core.Events.post_event(name, event)

    def listen_event(self, name: typing.Optional[str], callback,
                     even_type: typing.Union[HABApp.core.events.AllEvents, typing.Any] = HABApp.core.events.AllEvents
                     ) -> HABApp.core.EventListener:
        """
        Register and event listener
        :param name: name to listen to or '' for all event names
        :param callback: callback
        :param even_type: class to only make a call on class instances
        :return: Instance of EventListener
        """
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        listener = HABApp.core.EventListener(name, cb, even_type)
        self.__event_listener.append(listener)
        HABApp.core.Events.add_listener(listener)
        return listener

    def run_every(self, time: TYPING_TIME, interval, callback, *args, **kwargs) -> ScheduledCallback:
        """
        Run a function every interval
        :param date_time:
        :param interval:
        :param callback:
        :param args:
        :param kwargs:
        :return:
        """
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = ReoccurringScheduledCallback(time, interval, cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_on_day_of_week(self, time: TYPING_TIME, weekdays, callback, *args, **kwargs) -> ScheduledCallback:

        # names of weekdays in local language
        lookup = {datetime.date(2001, 1, i).strftime('%A'): i for i in range(1, 8)}
        lookup.update( {datetime.date(2001, 1, i).strftime('%A')[:3]: i for i in range(1, 8)})

        # abreviations in German and English
        lookup.update({"Mo": 1, "Di": 2, "Mi": 3, "Do": 4, "Fr": 5, "Sa": 6, "So": 7})
        lookup.update({"Mon": 1, "Tue": 2, "Wed": 3, "Thu": 4, "Fri": 5, "Sat": 6, "Sun": 7})
        lookup = {k.lower(): v for k, v in lookup.items()}

        if isinstance(weekdays, int) or isinstance(weekdays, str):
            weekdays = [weekdays]
        for i, val in enumerate(weekdays):
            if not isinstance(val, str):
                continue
            weekdays[i] = lookup[val.lower()]

        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = DayOfWeekScheduledCallback(time, weekdays, cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_on_every_day(self, time: TYPING_TIME, callback, *args, **kwargs) -> ScheduledCallback:
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = DayOfWeekScheduledCallback(time, [1, 2, 3, 4, 5, 6, 7], cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_on_workdays(self, time: TYPING_TIME, callback, *args, **kwargs) -> ScheduledCallback:
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = WorkdayScheduledCallback(time, cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_on_weekends(self, time: TYPING_TIME, callback, *args, **kwargs) -> ScheduledCallback:
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = WeekendScheduledCallback(time, cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_daily(self, callback, *args, **kwargs) -> ScheduledCallback:
        """
        Picks a random minute and second and runs the callback every hour
        :param callback:
        :param args:
        :param kwargs:
        :return:
        """
        start = datetime.timedelta(seconds=random.randint(0, 24 * 3600 - 1))
        interval = datetime.timedelta(days=1)
        return self.run_every(start, interval, callback, *args, **kwargs)

    def run_hourly(self, callback, *args, **kwargs) -> ScheduledCallback:
        """
        Picks a random minute and second and runs the callback every hour
        :param callback:
        :param args:
        :param kwargs:
        :return:
        """
        start = datetime.timedelta(seconds=random.randint(0, 3600 - 1))
        interval = datetime.timedelta(seconds=3600)
        return self.run_every(start, interval, callback, *args, **kwargs)

    def run_minutely(self, callback, *args, **kwargs) -> ScheduledCallback:
        start = datetime.timedelta(seconds=random.randint(0, 60 - 1))
        interval = datetime.timedelta(seconds=60)
        return self.run_every(start, interval, callback, *args, **kwargs)

    def run_at(self, date_time: TYPING_DATE_TIME, callback, *args, **kwargs) -> ScheduledCallback:
        "Run a function at a specified date_time"
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = ScheduledCallback(date_time, cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_in(self, seconds: int, callback, *args, **kwargs) -> ScheduledCallback:
        """Run a function in x seconds"""
        assert isinstance(seconds, int), f'{seconds} ({type(seconds)})'

        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = ScheduledCallback(datetime.timedelta(seconds=seconds), cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_soon(self, callback, *args, **kwargs) -> ScheduledCallback:
        """
        Run the callback as soon as possible (typically in the next second).
        :param callback:    function to call
        :param args:    args for the callback
        :param kwargs:  kwargs for the callback
        :return:
        """
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = ScheduledCallback( None, cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def get_rule_parameter(self, file_name: str, *keys) -> RuleParameter:
        assert isinstance(file_name, str), type(file_name)
        return RuleParameter(self.__runtime, file_name, *keys)

    def get_rule(self, rule_name: str):  # todo: einkommentieren mit python3.7 -> Rule:
        assert isinstance(rule_name, str), type(rule_name)
        return self.__runtime.rule_manager.get_rule(rule_name)

    # -----------------------------------------------------------------------------------------------------------------
    # deprecated stuff
    # -----------------------------------------------------------------------------------------------------------------
    def post_update(self, name, value):
        log.warning('self.post_update is deprecated! Use self.openhab.post_update or self.oh.post_update instead!')
        self.openhab.post_update(name, value)

    def send_command(self, name, value):
        log.warning('self.send_command is deprecated! Use self.openhab.send_command or self.oh.send_command instead!')
        self.openhab.send_command(name, value)

    def create_openhab_item(self, item_type, item_name, label="", category="", tags=[], groups=[]):
        log.warning('self.create_openhab_item is deprecated!'
                    'Use self.openhab.create_item or self.oh.create_item instead!')
        return self.openhab.create_item(item_type, item_name, label, category, tags, groups)


    def remove_openhab_item(self, item_name: str):
        log.warning('self.remove_openhab_item is deprecated!'
                    'Use self.openhab.remove_item or self.oh.remove_item instead!')
        return self.openhab.remove_item(item_name)

    def mqtt_publish(self, topic: str, payload, qos=None, retain=None):
        log.warning('self.mqtt_publish is deprecated! Use self.mqtt.publish instead!')
        return self.mqtt.publish(topic, payload, qos, retain)

    # -----------------------------------------------------------------------------------------------------------------
    # internal functions
    # -----------------------------------------------------------------------------------------------------------------
    def __get_rule_name(self, callback):
        return f'{self.rule_name}.{callback.__name__}' if self.rule_name else None

    @HABApp.util.PrintException
    def _check_rule(self):

        # Check if items do exists
        if not HABApp.core.Items.items:
            return None

        for listener in self.__event_listener:
            # Listener listens to all changes
            if listener.name is None:
                continue

            # check if specific item exists
            if not HABApp.core.Items.item_exists(listener.name):
                log.warning(f'Item "{listener.name}" does not exist (yet)! '
                            f'self.listen_event in "{self.rule_name}" may not work as intended.')

        for item in self.__watched_items:
            if not HABApp.core.Items.item_exists(item.name):
                log.warning(f'Item "{item.name}" does not exist (yet)! '
                            f'self.item_watch in "{self.rule_name}" may not work as intended.')

    @HABApp.util.PrintException
    def _process_events(self, now):

        # watch items
        clean_items = False
        for item in self.__watched_items:
            item.check(now)
            if item.is_canceled:
                clean_items = True
        if clean_items:
            self.__watched_items = [k for k in self.__watched_items if not k.is_canceled]

        # sheduled events
        clean_events = False
        for future_event in self.__future_events:  # type: ScheduledCallback
            future_event.check_due(now)
            future_event.execute()
            if future_event.is_finished:
                clean_events = True

        # remove finished events
        if clean_events:
            self.__future_events = [k for k in self.__future_events if not k.is_finished]
        return None

    @HABApp.util.PrintException
    def _cleanup(self):

        # Important: set the dicts to None so we don't schedule a future event during _cleanup
        # If dict is set to None we will crash instead but it is no problem because everything gets unloaded anyhow
        event_listeners = self.__event_listener
        future_events = self.__future_events

        self.__event_listener = None
        self.__future_events = None
        self.__watched_items = None

        # Actually remove the listeners/events
        for listener in event_listeners:
            HABApp.core.Events.remove_listener(listener)

        for event in future_events:
            event.cancel()



import concurrent.futures
import logging
import time
import traceback

default_logger = logging.getLogger('HABApp.Worker')


class WrappedFunction:
    _WORKERS = concurrent.futures.ThreadPoolExecutor(10, 'HabApp_')

    def __init__(self, func, logger=None, warn_too_long = True, name = None):
        assert callable(func)
        self._func = func

        # name of the function
        self.name = self._func.__name__ if not name else name

        self.__submitted = 0.0

        # Allow custom logger
        self.log = default_logger
        if logger:
            assert isinstance(logger, logging.getLoggerClass())
            self.log = logger

        self.__warn_too_long = warn_too_long


    def submit(self, *args, **kwargs):
        self.__submitted = time.time()
        WrappedFunction._WORKERS.submit(self.__execute, *args, **kwargs)

    def __execute(self, *args, **kwargs):

        __start = time.time()

        # notify if we don't process quickly
        if __start - self.__submitted > 0.05:
            self.log.warning(f'Starting of {self.name} took too long: {__start - self.__submitted:.1f}s. '
                             f'Maybe there are not enough threads?')

        # Execute the function
        try:
            self._func(*args, **kwargs)
        except Exception as e:
            self.log.error("{}\n{}".format(e, traceback.format_exc()))

        # log warning if execution takes too long
        __dur = time.time() - __start
        if self.__warn_too_long and __dur > 0.8:
            self.log.warning(f'Execution of {self.name} took too long: {__dur:.1f}s')



import datetime

from HABApp.classes import Color


def map_event_types(openhab_type: str, openhab_value: str):
    assert isinstance(openhab_type, str), type(openhab_type)
    assert isinstance(openhab_value, str), type(openhab_value)

    if openhab_type == 'UnDef' or openhab_value == 'NULL':
        return None

    if openhab_type == "Number":
        return int(openhab_value)

    if openhab_type == 'Percent':
        return float(openhab_value)

    if openhab_type == "Decimal":
        try:
            return int(openhab_value)
        except ValueError:
            return float(openhab_value)

    if openhab_type == "DateTime":
        # 2018-11-19T09:47:38.284+0100
        return datetime.datetime.strptime(openhab_value.replace('+', '000+'), '%Y-%m-%dT%H:%M:%S.%f%z')

    if openhab_type == "HSB":
        return Color(*[float(k) for k in openhab_value.split(',')])

    return openhab_value



import threading


class Counter:
    def __init__(self, initial_value=0, on_change=None):
        assert isinstance(initial_value, int)

        self.__lock = threading.Lock()

        self.__initial_value = initial_value
        self.__value = self.__initial_value

        # func which gets called when the counter changes
        self.on_change = on_change
        if self.on_change:
            self.on_change(self.__value)

    @property
    def value(self) -> int:
        return self.__value

    def reset(self):
        with self.__lock:
            changed = self.__value != self.__initial_value
            self.__value = self.__initial_value
            value = self.__value

        if changed and self.on_change:
            self.on_change(value)

    def increase(self, step=1):
        with self.__lock:
            self.__value += step
            value = self.__value

        if self.on_change:
            self.on_change(value)

    def decrease(self, step=1):
        with self.__lock:
            self.__value -= step
            value = self.__value

        if self.on_change:
            self.on_change(value)



import pathlib

from watchdog.observers import Observer

from .fileeventhandler import SimpleFileEventHandler


class FolderWatcher:
    def __init__(self):

        self.__observer = Observer()
        self.__folders = {}

        self.__started = False


    def start(self, shutdown_helper):
        from ..shutdown_helper import ShutdownHelper
        assert isinstance(shutdown_helper, ShutdownHelper)

        # we shall only start once!
        assert self.__started is False
        self.__started = True

        # start watching the folders
        self.__observer.start()

        # register for proper shutdown
        shutdown_helper.register_func(self.__observer.stop)
        shutdown_helper.register_func(self.__observer.join, last=True)
        return None

    def watch_folder(self, folder: pathlib.Path, file_ending: str, event_target,
                     watch_subfolders = False, worker_factory=None):
        assert isinstance(folder, pathlib.Path)
        assert folder.is_dir()

        folder = str(folder)
        assert folder not in self.__folders, folder

        handler = SimpleFileEventHandler(event_target=event_target, file_ending=file_ending,
                                         worker_factory=worker_factory)
        self.__folders[folder] = self.__observer.schedule(handler, path=folder, recursive=watch_subfolders)

    def unwatch_folder(self, folder):
        if isinstance(folder, pathlib.Path):
            folder = str(folder)
        assert isinstance(folder, str), type(folder)

        self.__observer.unschedule(self.__folders.pop(folder))
